#include <avr/io.h>
#include <util/delay.h>

#include <avr/interrupt.h>

#include <stdbool.h>
#include <stdio.h>
#include <string.h>

#include "usart.h"
#include "spi.h"
#include "pff.h"
#include "adc.h"

/*
 * Four-Character Code - folosit pentru a indentifica formate de date
 */
#define FCC(c1, c2, c3, c4) \
	(((DWORD)(c4) << 24) + \
	 ((DWORD)(c3) << 16) + \
	 (( WORD)(c2) <<  8) + \
	 (( BYTE)(c1) <<  0))

uint8_t	buf[2][256];	// wave output buffers (double buffering)
const	 uint16_t	buf_size = 256;	// front and back buffer sizes
volatile uint8_t	buf_front = 0;	// front buffer index (current buffer used)
volatile uint8_t	buf_pos = 0;	// current buffer position
volatile uint8_t	buf_sync = 0;

#define BUF_FRONT	(buf[buf_front])
#define BUF_BACK	(buf[1 - buf_front])

#define ROWS 4
#define COLUMNS 4

FATFS fs;					// sistemul de fisiere
uint8_t rec;
int count = 0;
WORD ret;
int rp = 0;
char filename[32] = { 0 };

static int nr_of_files = 13;
char filenames[25][20] = {
	"S1",
	"S2",
	"S3",
	"S5",
	"S6",
	"S7",
	"S9",
	"S10",
	"S11",
	"S13",
	"S14",
	"S15",
	"record"  // this is the test file
};

static int rows[] = {
	PC4,
	PC2,
	PC1,
	PC0
};

static int columns[] = {
	PC3,
	PC5,
	PC6,
	PA7
};

ISR(ADC_vect)
{
	rec = ADC;
	FRESULT b = pf_write(&rec, 1, &ret);

	if(ret == 0){
		printf("err: %d\r\n", b);
	}

	count++;
	if(count == 1048576) {
		printf("reached EOF\r\n");
		cli();
	}
}

ISR(TIMER0_COMPA_vect)
{
	if(rp == 2) {
		// start conversion
		ADCSRA |= (1 << ADSC);
	} else if (rp == 1) {
		OCR1B = BUF_FRONT[buf_pos++];
		// swap buffers when end is reached (end is 256 <=> overflow to 0)
		if(buf_pos == 0)
			buf_front = 1 - buf_front;
	}
}

void timer0_start_record(void)
{
	// interrupt on compare A
	TIMSK0 |= (1 << OCIE0A);
	// CTC, top OCRA
	TCCR0A |= (1 << WGM01);

	// prescaler 8
	TCCR0B |= (2 << CS00);

	// create interrupt @ 16KHz
	OCR0A = 0x7D;
}

void timer0_start_play_record(void)
{
	// interrupt on compare A
	TIMSK0 |= (1 << OCIE0A);

	// CTC, top OCRA
	TCCR0A |= (1 << WGM01);
	
	// prescaler 8, ocr0a = 250 => 8khz
	TCCR0B |= (2 << CS00);
	OCR0A = 0xFA;
}

void timer0_start_play(void)
{
	// interrupt on compare A
	TIMSK0 |= (1 << OCIE0A);
	// CTC, top OCRA
	TCCR0B |= (0 << WGM02);
	TCCR0A |= (1 << WGM01) | (0 << WGM00);
	// prescaler 8
	TCCR0B |= (2 << CS00);
}

void timer0_stop(void)
{
	TCCR0B = 0;
	TCCR0A = 0;
	TIMSK0 = 0;
	OCR0A = 0;
	TCNT0 = 0;
}

void timer1_start(void)
{
	// 8-bit FastPWM
	TCCR1B |= (1 << WGM12);
	TCCR1A |= (1 << WGM10);
	// channel B inverted
	TCCR1A |= (1 << COM1B0) | (1 << COM1B1);
	// prescaler 1
	TCCR1B |= (1 << CS10);
}

void timer1_stop(void)
{
	TCCR1B = 0;
	TCCR1A = 0;
	OCR1B = 0;
	TCNT1 = 0;
}

void play_record(char *filename)
{
	WORD bytes_read;

	char fname[25];

	memset(fname, 0, 25);
	strncpy(fname, filename, strlen(filename));
	strncat(fname, ".wav", 4);
	printf("Name is: %s\r\n", fname);
	// wait for succesful open
	while(1) {
		FRESULT x = pf_open(fname);
		if(x == 0){
			break;
		}
		_delay_ms(10);
	}
	
	printf("succes\r\n");

	unsigned int size;
	pf_read(&size, 4, &bytes_read);
	// fill front buffer
	ret = pf_read(BUF_FRONT, buf_size, &bytes_read);
	if(ret != FR_OK)
		return;
	if(bytes_read < buf_size)
		return;

	// reset front buffer index
	buf_pos = 0;

	// start output
	timer0_start_play_record();
	timer1_start();
	DDRD |= (1 << PD4);

	printf("record playing...\r\n");

	while(1) {
		uint8_t old_buf_front = buf_front;
		
		// fill back buffer
		ret = pf_read(BUF_BACK, buf_size, &bytes_read);
		if(ret != FR_OK)
			break;
		if(bytes_read < buf_size)
			break;

		// wait for buffer swap
		while(old_buf_front == buf_front) ;
	}
	
	// stop output
	DDRD &= ~(1 << PD4);
	timer1_stop();
	timer0_stop();
	
	printf("done\r\n");
}

void recording(char *filename)
{
	int working = 0;
	
	char fname[25];

	memset(fname, 0, 25);
	strncpy(fname, filename, strlen(filename));
	strncat(fname, ".wav", 4);
	printf("Name is: %s\r\n", fname);

	// wait for succesful open
	while(1) {
		FRESULT x = pf_open(fname);
		if(x == 0){
			break;
		}
		_delay_ms(10);
	}

	printf("succes\r\n");
	_delay_ms(1000);

	while((PINB & (1 << PB2)) == 0)
		if(working == 0) {
			printf("recording...\r\n");

			// porneste timerul
			timer0_start_record();
			working = 1;
		}
	if(working == 1) {
		// opreste timerul si scrie numarul de bytes srisi
		timer0_stop();
		working = 0;
		
		_delay_ms(10);
		
		pf_write(0, 0, &ret);
		
		ret = pf_lseek(0);
		if(ret != FR_OK) {
			printf("%lu\r\n", fs.fptr);
			while(1);
		}
		ret = pf_write(&count, 4, &ret);
		if(ret != FR_OK) {
			printf("1\r\n");
			while(1);
		}
		
		count = 0;
		
		printf("done recording\r\n");
	}
}

void mount()
{
	USART0_print("begin mounting...\r\n");
	while(pf_mount(&fs) != FR_OK)
	{
		USART0_print("mounting...\r\n");
		// wait a while and retry
		_delay_ms(1000);
	}
	_delay_ms(1000);
	USART0_print("mounted\r\n");
}

static char const *get_switch(int row, int column)
{
	// primul rand de butoane
	if (row == 0 && column == 0)
		return "S1";
	if (row == 0 && column == 1)
		return "S2";
	if (row == 0 && column == 2)
		return "S3";
	if (row == 0 && column == 3)
		return "S4";

	// al doilea rand de butoane
	if (row == 1 && column == 0)
		return "S5";
	if (row == 1 && column == 1)
		return "S6";
	if (row == 1 && column == 2)
		return "S7";
	if (row == 1 && column == 3)
		return "S8";

	// al treilea rand de butoane
	if (row == 2 && column == 0)
		return "S9";
	if (row == 2 && column == 1)
		return "S10";
	if (row == 2 && column == 2)
		return "S11";
	if (row == 2 && column == 3)
		return "S12";

	// al patrulea rand de butoane
	if (row == 3 && column == 0)
		return "S13";
	if (row == 3 && column == 1)
		return "S14";
	if (row == 3 && column == 2)
		return "S15";
	if (row == 3 && column == 3)
		return "S16";

	return NULL;
}

int main(void)
{
	int i, j;
	int file_no = 12;
	int ch_file = 0;
	char record_file[25] = "record";

	// inits
	USART0_init();
	ADC_init();
	mount();
	sei();

	// configurarea stdout
	FILE mystdout;
	mystdout.put = USART0_printf;
	mystdout.flags = __SWR;
	stdout = &mystdout;

	DDRC &= ~(1 << PC3);
	PORTC |= (1 << PC3);

	DDRC &= ~(1 << PC5);
	PORTC |= (1 << PC5);

	DDRC &= ~(1 << PC6);
	PORTC |= (1 << PC6);

	DDRA &= ~(1 << PA7);
	PORTA |= (1 << PA7);

	DDRB &= ~(1 << PB2);
	PORTB |= (1 << PB2);

	DDRA |= (1 << PA2); // senzor sunet : input

	printf("\033\143");

	while(1) {
		// verificare pentru a se efectua o singura operatie odata
		if(rp == 0) {
			if((PINB & (1 << PB2)) == 0) {
				printf("Before record\n\r");
				rp = 2;
				recording(filenames[file_no]);
				rp = 0;
				_delay_ms(100);
			}
			for (i = 0; i < ROWS; ++i) {
				DDRC |= (1 << rows[i]);
				for (j = 0; j < COLUMNS - 1; ++j) {
					if ((PINC & (1 << columns[j])) == 0) {
						rp = 1;
						printf("%s is pressed!\r\n",  get_switch(i, j));
						play_record((char *)get_switch(i, j));
						rp = 0;
						_delay_ms(100);
					}
				}
				if ((PINA & (1 << columns[3])) == 0) {
					/* choose filename to be written */
					if (!strcmp(get_switch(i, 3), "S4")) {
						ch_file ^= 1;
						if (ch_file == 0) {
							printf("\033\143");
						}
						_delay_ms(100);
					}
					if (!strcmp(get_switch(i, 3), "S16")) {
						rp = 1;
						printf("%s is pressed!\r\n",  get_switch(i, 3));
						play_record(record_file);
						rp = 0;
						_delay_ms(100);
					}
					if (!strcmp(get_switch(i, 3), "S8")) {
						file_no = (file_no + 1) % nr_of_files;
						_delay_ms(100);
					}
					if (!strcmp(get_switch(i, 3), "S12")) {
						file_no--;
						if (file_no < 0) {
							file_no = nr_of_files - 1;
						}
						_delay_ms(100);
					}
				}
				DDRC &= ~(1 << rows[i]);
			}
			_delay_ms(100);
			if (ch_file == 1) {
				printf("\033\143");
				printf("Choose file\r\n");
				printf("Current file: %s\r\n", filenames[file_no]);
			}
		}
	}

	return 0;
}
